//---------------------------------------------------------------------------

#include <vcl.h>

#pragma hdrstop

#include "ComPort.h"
#pragma package(smart_init)

//---------------------------------------------------------------------------
// ValidCtrCheck is used to assure that the components created do not have
// any pure virtual functions.
//

static inline void ValidCtrCheck(MComPort *)
{
	new MComPort(NULL);
}
//---------------------------------------------------------------------------
namespace Comport
{
	void __fastcall PACKAGE Register()
	{
		TComponentClass classes[1] = {__classid(MComPort)};
		RegisterComponents("Lyncon", classes, 0);
	}
}
//---------------------------------------------------------------------------
__fastcall MComPort::MComPort(TComponent* Owner) : TComponent(Owner) {
    // Component initialization
	Handle = INVALID_HANDLE_VALUE;
    FThread = NULL;
    // Fixed port values
    Setup.DCBlength = sizeof( DCB );
    Setup.fBinary = true;
	Setup.wReserved = 0;
    // User and developer changeable port values:
    BaudRate    = bd9600;
    ByteSize    = bs8;
    Parity      = paNo;
    StopBits    = sb1_0;
    XonChar     = 'Q'-64;       // ^Q
    XoffChar    = 'S'-64;       // ^S
}
__fastcall MComPort::~MComPort() {
    Close();
}

HANDLE __fastcall MComPort::Open() {
	if( Handle == INVALID_HANDLE_VALUE ) {
        try {
			FThread = new MComPortThread(this);
            FThread->Resume();
        }
        catch( ... ) {
			Handle = INVALID_HANDLE_VALUE;
			if( OnError ) OnError( this, GetLastError() );
        }
    }
    return Handle;
}

bool __fastcall MComPort::Close() {
	bool retval=false;

	if( Handle != INVALID_HANDLE_VALUE ) {
		FThread->Close();
		retval = true;
    }
    return retval;
}
DWORD __fastcall MComPort::Read( LPVOID lpBuffer, DWORD nNumberOfBytesToRead ) {
    DWORD lNumberOfBytesRead=-1;
    OVERLAPPED osStatus; //should probably be the OVERLAPPed struct in the Thread?
    if( ReadFile(
        Handle,
        lpBuffer,
        nNumberOfBytesToRead,
        &lNumberOfBytesRead,
        &osStatus ) )
    {

    }
    return lNumberOfBytesRead;
}
DWORD __fastcall MComPort::Write(LPVOID lpBuffer, DWORD nNumberOfBytesToWrite ) {
    DWORD lNumberOfBytesWritten=-1;
    OVERLAPPED osStatus; //should probably be the OVERLAPPed struct in the Thread?
    if( WriteFile(
        Handle,
        lpBuffer,
        nNumberOfBytesToWrite,
        &lNumberOfBytesWritten,
        &osStatus ) )
    {

    }
    return lNumberOfBytesWritten;
}

/*
**                            The worker thread
*/

__fastcall MComPortThread::MComPortThread(MComPort *ComPort) : TThread( true )
{
    FComPort = ComPort;
    FreeOnTerminate = true;           // destroy thread when Execute() finishes

    FComPort->Handle = CreateFile(
        FComPort->PortName.c_str(),   // Port path, "\\.\COM2"...
        GENERIC_READ | GENERIC_WRITE, // access modes
        0,                            // share mode
        NULL,                         // pointer to security attributes
        OPEN_EXISTING,                // how to create
        FILE_FLAG_OVERLAPPED,         // file attributes
        0                             // handle to file with attributes to copy
    );

	if( FComPort->Handle == INVALID_HANDLE_VALUE ) {

        if( FComPort->OnError )
            FComPort->OnError( FComPort, GetLastError() );// Generate an OnError event

        Terminate(); // Mark thread for termination
    }
}
//---------------------------------------------------------------------------
__fastcall MComPortThread::~MComPortThread() {
}

//---------------------------------------------------------------------------
void __fastcall MComPortThread::Execute() {
    if( !Terminated ) {     // Investigate if this is really needed
		if( SetupPort() ) { // Misc setup stuff

            Synchronize( (TThreadMethod) &CallOpened ); // Generate an OnOpened event

            Work();                          // guess...

			CloseHandle( osStatus.hEvent );  // This handle created in SetupPort()
			CloseHandle( FComPort->Handle ); // Opened in the thread constructor
			FComPort->Handle = INVALID_HANDLE_VALUE;
			Synchronize( (TThreadMethod) &CallClosed ); // Generate an OnClosed event

		} else {
			// Port was never successfully setup and therefore no OnOpen was
			// sent, so we close it without generating an OnClosed event.
			// An OnError event should have been generated in SetupPort()
			// instead.
			CloseHandle( FComPort->Handle );
			FComPort->Handle = INVALID_HANDLE_VALUE;
        }
        FComPort->Handle = INVALID_HANDLE_VALUE;
    }
}

bool __fastcall MComPortThread::SetupPort() {
    bool retval=false;

	osStatus.hEvent = CreateEvent(NULL, TRUE, FALSE, NULL); // To be used in WaitCommEvent() calls

	if( osStatus.hEvent ) {

        if( FComPort->SetState() ) {

			retval = SetCommMask( FComPort->Handle,
                (FComPort->OnBreak  ? EV_BREAK  :0) | \
                (FComPort->OnCts    ? EV_CTS    :0) | \
                (FComPort->OnDsr    ? EV_DSR    :0) | \
                (FComPort->OnErr    ? EV_ERR    :0) | \
                (FComPort->OnRing   ? EV_RING   :0) | \
                (FComPort->OnRlsd   ? EV_RLSD   :0) | \
                (FComPort->OnRxchar ? EV_RXCHAR :0) | \
                (FComPort->OnRxflag ? EV_RXFLAG :0) | \
                (FComPort->OnTxempty? EV_TXEMPTY:0) );
/* Other events declared in winbase.h
#define EV_PERR             0x0200  // Printer error occured
#define EV_RX80FULL         0x0400  // Receive buffer is 80 percent full
#define EV_EVENT1           0x0800  // Provider specific event 1
#define EV_EVENT2           0x1000  // Provider specific event 2
*/
        }
        if( !retval ) {
            SyncError();
			CloseHandle( osStatus.hEvent );
        }
    } else SyncError();

    return retval;
}
void __fastcall MComPortThread::Close() {
	Terminate();
	SetEvent( osStatus.hEvent ); // theoretically, it may be closed already ...
}
void __fastcall MComPortThread::Work() {
    BOOL       fWaitingOnStat = FALSE;
    DWORD      dwRes;
    DWORD      dwOvRes;

  while( !Terminated ) {
    if( !fWaitingOnStat ) {
		if( !WaitCommEvent( FComPort->Handle, &dwCommEvent, &osStatus) ) {
            if (GetLastError() == ERROR_IO_PENDING) {
                fWaitingOnStat = TRUE;
            } else {
                // error in WaitCommEvent; abort
                Terminate();
            }
        } else {
            // WaitCommEvent returned immediately.
            // Deal with status event as appropriate.
            Synchronize( (TThreadMethod) &CallEvent );
        }
    }

    if( fWaitingOnStat ) {
		// Wait for an event to occur.
		dwRes = WaitForSingleObject(osStatus.hEvent, INFINITE );
		
        switch(dwRes) { // Event occurred.
        case WAIT_OBJECT_0:
			if( !GetOverlappedResult( FComPort->Handle, &osStatus, &dwOvRes, FALSE) ) {
                // An error occurred in the overlapped operation;
                // call GetLastError to find out what it was
				// and abort if it is fatal.
				if( GetLastError() != ERROR_IO_PENDING ) Terminate(); // not likely when GetOverlappedResult() is called with bWait=FALSE

            } else {
                // Status event is stored in the event flag
                // specified in the original WaitCommEvent call.
                // Deal with the status event as appropriate.
                Synchronize( (TThreadMethod) &CallEvent );
            }
            // Set fWaitingOnStat flag to indicate that a new
            // WaitCommEvent is to be issued.
            fWaitingOnStat = FALSE;
            break;

		case WAIT_TIMEOUT: // not really going to happen - waiting INFINITE
			// Operation isn't complete yet. fWaitingOnStat flag
			// isn't changed since I'll loop back around and I don't want
			// to issue another WaitCommEvent until the first one finishes.
			break;

        // case WAIT_FAILED:
        default:
            // Error in the WaitForSingleObject; abort
            // This indicates a problem with the OVERLAPPED structure's
            // event handle.
            Terminate();
        }
    }
  }
}

void __fastcall MComPortThread::CallEvent() {
    if( dwCommEvent & EV_BREAK )   FComPort->OnBreak( FComPort );
    if( dwCommEvent & EV_CTS )     FComPort->OnCts( FComPort );
    if( dwCommEvent & EV_DSR )     FComPort->OnDsr( FComPort );
    if( dwCommEvent & EV_ERR )     FComPort->OnDsr( FComPort );
    if( dwCommEvent & EV_RING )    FComPort->OnRing( FComPort );
    if( dwCommEvent & EV_RLSD )    FComPort->OnRlsd( FComPort );
    if( dwCommEvent & EV_RXCHAR )  FComPort->OnRxchar( FComPort );
    if( dwCommEvent & EV_RXFLAG )  FComPort->OnRxflag( FComPort );
    if( dwCommEvent & EV_TXEMPTY ) FComPort->OnTxempty( FComPort );
}

void __fastcall MComPortThread::CallOpened() {
	if( FComPort->OnOpened ) FComPort->OnOpened( FComPort );
}

void __fastcall MComPortThread::CallClosed() {
	if( FComPort->OnClosed ) FComPort->OnClosed( FComPort );
}

inline void __fastcall MComPortThread::SyncError() {
	ErrorCode = GetLastError();
	Synchronize( (TThreadMethod) &CallError );
}

void __fastcall MComPortThread::CallError() {
    if( FComPort->OnError ) FComPort->OnError( FComPort, ErrorCode );
}

//---------------------------------------------------------------------------
