//---------------------------------------------------------------------------

#ifndef ComPortH
#define ComPortH
//---------------------------------------------------------------------------
#include <SysUtils.hpp>
#include <Classes.hpp>
//---------------------------------------------------------------------------
enum Baud {     bd110=110,      bd300=300,      bd600=600,
                bd1200=1200,    bd2400=2400,    bd4800=4800,
                bd9600=9600,    bd14400=14400,  bd19200=19200,
                bd38400=38400,  bd56000=56000,  bd57600=57600,
                bd115200=115200,bd128000=128000,bd256000=256000 };

enum paType { paNo, paOdd, paEven, paMark, paSpace };
enum Stopbits { sb1_0, sb1_5, sb2_0 };
enum DtrControl { dtrDisable, dtrEnable, dtrHandshake };
enum RtsControl { rtsDisable, rtsEnable, rtsHandshake, rtsToggle };
enum bsType { bs4, bs5, bs6, bs7, bs8 };

class MComPortThread; // Forward declaration

typedef void __fastcall (__closure *FOnError)( TObject *Sender, DWORD ErrorCode );
typedef void __fastcall (__closure *FOnOpened)( TObject *Sender );
typedef void __fastcall (__closure *FOnClosed)( TObject *Sender );

class PACKAGE MComPort : public TComponent
{
private:
        MComPortThread *FThread;

        AnsiString FPortName;
        // Events
        FOnError FError;
        FOnOpened FOpened;
        FOnOpened FClosed;
        TNotifyEvent FBreak,FCts,FDsr,FErr,FRing,FRlsd,FRxchar,FRxflag,FTxempty;
protected:
public:
        HANDLE Handle;
		DCB Setup;

        __fastcall MComPort(TComponent* Owner);
        __fastcall ~MComPort();
        HANDLE __fastcall Open();
        bool __fastcall Close();
        DWORD __fastcall Read( LPVOID lpBuffer, DWORD nNumberOfBytesToRead );
        DWORD __fastcall Write(LPVOID lpBuffer, DWORD nNumberOfBytesToWrite );
        bool __fastcall EscapeFunction( DWORD Function ) { return EscapeCommFunction( Handle, Function ); }
        bool __fastcall SetState() { return SetCommState( Handle, &Setup ); }

        //- published properties get/set methods
        Baud __fastcall GetBaudRate() { return (Baud) Setup.BaudRate; }
        void __fastcall SetBaudRate( Baud x ) { Setup.BaudRate = x; }

        bool __fastcall GetfParity() { return Setup.fParity; };
        void __fastcall SetfParity( bool x ) { Setup.fParity = x; }

        bool __fastcall GetfOutxCtsFlow() { return Setup.fOutxCtsFlow; }
        void __fastcall SetfOutxCtsFlow( bool x ) { Setup.fOutxCtsFlow = x; }

        bool __fastcall GetfOutxDsrFlow() { return Setup.fOutxDsrFlow; }
        void __fastcall SetfOutxDsrFlow( bool x ) { Setup.fOutxDsrFlow = x; }

        DtrControl __fastcall GetfDtrControl() { return (DtrControl) Setup.fDtrControl; }
        void __fastcall SetfDtrControl( DtrControl x ) { Setup.fDtrControl = x; }

        bool __fastcall GetfDsrSensitivity() { return Setup.fDsrSensitivity; }
        void __fastcall SetfDsrSensitivity( bool x ) { Setup.fDsrSensitivity = x; }

        bool __fastcall GetfTXContinueOnXoff() { return Setup.fTXContinueOnXoff; }
        void __fastcall SetfTXContinueOnXoff( bool x ) { Setup.fTXContinueOnXoff = x; }

        bool __fastcall GetfOutX() { return Setup.fOutX; }
        void __fastcall SetfOutX( bool x ) { Setup.fOutX = x; }

        bool __fastcall GetfInX() { return Setup.fInX; }
        void __fastcall SetfInX( bool x ) { Setup.fInX = x; }

        bool __fastcall GetfErrorChar() { return Setup.fErrorChar; }
        void __fastcall SetfErrorChar( bool x ) { Setup.fErrorChar = x; }

        bool __fastcall GetfNull() { return Setup.fNull; }
        void __fastcall SetfNull( bool x ) { Setup.fNull = x; }

        RtsControl __fastcall GetfRtsControl() { return (RtsControl) Setup.fRtsControl; }
        void __fastcall SetfRtsControl( RtsControl x ) { Setup.fRtsControl = x; }

        bool __fastcall GetfAbortOnError() { return Setup.fAbortOnError; }
        void __fastcall SetfAbortOnError( bool x ) { Setup.fAbortOnError = x; }

        bsType __fastcall GetByteSize() { return (bsType) (Setup.ByteSize-4); }
        void __fastcall SetByteSize( bsType x ) { Setup.ByteSize = (BYTE) (x+4); }

        paType __fastcall GetParity() { return (paType) Setup.Parity; }
        void __fastcall SetParity( paType x ) { Setup.Parity = x; }

        Stopbits __fastcall GetStopBits() { return (Stopbits) Setup.StopBits; }
        void __fastcall SetStopBits( Stopbits x ) { Setup.StopBits = x; }

__published:
    __property AnsiString PortName = { read=FPortName, write=FPortName };

    __property Baud BaudRate = { read=GetBaudRate, write=SetBaudRate };                                 // current baud rate
    __property bool fParity = { read=GetfParity, write=SetfParity, default=true };                      //1b enable parity checking
    __property bool fOutxCtsFlow = { read=GetfOutxCtsFlow, write=SetfOutxCtsFlow };                     //1b CTS output flow control
    __property bool fOutxDsrFlow = { read=GetfOutxDsrFlow, write=SetfOutxDsrFlow };                     //1b DSR output flow control
    __property DtrControl fDtrControl = { read=GetfDtrControl, write=SetfDtrControl };                  //2b DTR flow control type
    __property bool fDsrSensitivity = { read=GetfDsrSensitivity, write=SetfDsrSensitivity };            //1b DSR sensitivity

    __property bool fTXContinueOnXoff = { read=GetfTXContinueOnXoff, write=SetfTXContinueOnXoff };      //1b XOFF continues Tx
    __property bool fOutX = { read=GetfOutX, write=SetfOutX };                                          //1b XON/XOFF out flow control
    __property bool fInX = { read=GetfInX, write=SetfInX };                                             //1b XON/XOFF in flow control
    __property bool fErrorChar = { read=GetfErrorChar, write=SetfErrorChar };                           //1b enable error replacement
    __property bool fNull = { read=GetfNull, write=SetfNull };                                          //1b enable null stripping
    __property RtsControl fRtsControl = { read=GetfRtsControl, write=SetfRtsControl };                  //2b RTS flow control
    __property bool fAbortOnError = { read=GetfAbortOnError, write=SetfAbortOnError };                  //1b abort reads/writes on error

    __property WORD XonLim = { read=Setup.XonLim, write=Setup.XonLim };                                 // transmit XON threshold
    __property WORD XoffLim = { read=Setup.XoffLim, write=Setup.XoffLim };                              // transmit XOFF threshold
    __property bsType ByteSize = { read=GetByteSize, write=SetByteSize,default=bs8 };                   // number of bits/byte, 4-8
    __property paType Parity = { read=GetParity, write=SetParity };                                     // 0-4=no,odd,even,mark,space
    __property Stopbits StopBits = { read=GetStopBits, write=SetStopBits };                             // 0,1,2 = 1, 1.5, 2
    __property char XonChar = { read=Setup.XonChar, write=Setup.XonChar };                              // Tx and Rx XON character
    __property char XoffChar = { read=Setup.XoffChar, write=Setup.XoffChar };                           // Tx and Rx XOFF character
    __property char ErrorChar = { read=Setup.ErrorChar, write=Setup.ErrorChar };                        // error replacement character

    __property char EofChar = { read=Setup.EofChar, write=Setup.EofChar };                              // end of input character
    __property char EvtChar = { read=Setup.EvtChar, write=Setup.EvtChar };                              // received event character

    // Events - Application
    __property FOnOpened OnOpened = { read=FOpened, write=FOpened }; // Called just after port has been opened
    __property FOnClosed OnClosed = { read=FClosed, write=FClosed }; // Called just after port has been closed
    __property FOnError OnError = { read=FError, write=FError };
    // Events - Status changes while port is open
    __property TNotifyEvent OnBreak = { read=FBreak, write=FBreak };
    __property TNotifyEvent OnCts = { read=FCts, write=FCts };
    __property TNotifyEvent OnDsr = { read=FDsr, write=FDsr };
    __property TNotifyEvent OnErr = { read=FErr, write=FErr };
    __property TNotifyEvent OnRing = { read=FRing, write=FRing };
    __property TNotifyEvent OnRlsd = { read=FRlsd, write=FRlsd };
    __property TNotifyEvent OnRxchar = { read=FRxchar, write=FRxchar };
    __property TNotifyEvent OnRxflag = { read=FRxflag, write=FRxflag };
    __property TNotifyEvent OnTxempty = { read=FTxempty, write=FTxempty };
};



class MComPortThread : public TThread
{
private:
        MComPort   *FComPort;
        OVERLAPPED osStatus;
		DWORD      ErrorCode;
		DWORD      dwCommEvent;
        void __fastcall CallOpened();
        void __fastcall CallClosed();
        inline void __fastcall SyncError(); // Use in subthread
        void __fastcall CallError(); // Use in main VCL thread
        void __fastcall CallEvent();
protected:
		void __fastcall Execute();
		bool __fastcall SetupPort();
		void __fastcall Work();
public:
		__fastcall MComPortThread( MComPort *ComPort );
		__fastcall ~MComPortThread();
		void __fastcall Close();
};

//---------------------------------------------------------------------------
#endif
